'''
Polynomial Approximation
'''
from smolyak.applications.polynomials.polynomial_approximation import PolynomialApproximator
from smolyak.applications.polynomials.probability_distributions import ProbabilityDistribution



'''
smolyak includes the following modules and packages:

* smolyak.py: General, adaptive and non-adaptive, implementations of Smolyak's algorithm
        in class SparseApproximator. 
         
* indices.py: Various tools for handling multi-indices, such as:
        * a sparse object oriented representation of multi-indices in class MultiIndex
        * the class MISet for connected multi-index sets
        * a class that computes mixed differences
        * various functions for the generation of multi-index sets in different shapes, such as
            simplex, hyperbolic_cross, pyramids, or determined by user-defined functions (get_admissible_indices)
        * computation of combination rule coefficients

* applications: Various numerical approximation algorithms that can be 
    accelerated using Smolyak's algorithm, including:
    * Simulation of particle systems
    * Solution of partial differential equations
    * Polynomial approximation
    
'''
from smolyak.smolyak import SparseApproximator
from smolyak import indices



import numpy as np
from smolyak.applications.particle_systems.particle_approximations import inverse_transform_sampling, \
    random_sampling

def univariate_integral(particles, phi, h, shift):
    N = len(particles)
    if N == 0:
        raise ValueError('Need at least one particle')
    if shift:
        phi_particles = phi(np.array(particles) + h)
    else:
        phi_particles = phi(np.array(particles))
    return 1. / N * np.sum(phi_particles) + h

def time_stepping(particles, phi, h, shift):
    N_steps = int(1. / h)
    h = 1. / N_steps
    N_particles = len(particles)
    for __ in range(N_steps):
        new_particles = particles.copy()
        for (i, particle) in enumerate(particles):
            for (j, particle_2) in enumerate(particles):
                if i != j:
                    new_particles[i] = new_particles[i] + 1. / N_particles * h * (particle_2 - particle) * np.power(np.absolute(particle_2 - particle), 3)
        particles = new_particles
    phi_particles = phi(new_particles)
    return 1. / N_particles * np.sum(phi_particles)
                    
    
    

def univariate_integral_approximation(N_particles, h, invcdf=None, pdf=None, bound=None, phi=None, shift=None, app_type='time_stepping'):
    if invcdf:
        particles = inverse_transform_sampling(invcdf, N_particles)
        
    else:
        particles = random_sampling(pdf, bound, N_particles)
    if app_type == 'artificial':
        return univariate_integral(particles, phi, h, shift)
    elif app_type == 'time_stepping':
        return time_stepping(particles, phi, h, shift)



'''
Orthogonal polynomials
'''
import numpy as np

def evaluate_orthonormal_polynomials(X, max_degree, measure, interval=(0, 1),derivative = 0):
    r'''
    Evaluate orthonormal polynomials in :math:`X`.
    The endpoints of `interval` can be specified when `measure` is uniform or Chebyshev.
    
    :param X: Locations of desired evaluations
    :type X:  One dimensional np.array
    :param max_degree: Maximal degree of polynomial basis.
    :param measure: Orthogonality measure. `u` for uniform, `c` for Chebyshev, `h` for Gauss/Hermite
    :param interval: Domain
    :type interval: tuple or list
    :rtype: numpy.array of size :code:`X.shape[0] x (max_degree+1)`
    '''
    max_degree = int(max_degree)
    if derivative and measure not in ['t','u','h']:
        raise ValueError('Derivative not implemented for Chebyshev polynomials')
    if derivative>1 and measure not in ['h','t']:
        raise ValueError('Second derivative only implemented for Taylor and Hermite polynomials')
    if measure in ['u', 'c']:
        Xtilde = (X - (interval[1] + interval[0]) / 2.) / ((interval[1] - interval[0]) / 2.) 
        if measure == 'u':
            y = legendre_polynomials(Xtilde, max_degree + 1)
            if derivative ==0:
                return y
            if derivative > 0:
                for _ in range(derivative):
                    y1 = np.zeros(y.shape)
                    y1[:,1] = np.sqrt(3)*np.ones(y.shape[0])
                    for j in range(2,max_degree+1):
                        y1[:,j] = np.sqrt(2*j+1)*((2*j-1)*y[:,j-1]/np.sqrt(2 * j - 1) + y1[:,j-2]/np.sqrt(2*j-3))
                    y=y1
                return (2/(interval[1]-interval[0]))**derivative*y
        elif measure == 'c':
            return chebyshev_polynomials(Xtilde, max_degree + 1)
    elif measure == 'h':
        Xtilde = X / interval[1]
        y = hermite_polynomials(Xtilde, max_degree + 1)
        factorial = np.ones((1,max_degree+1))
        for i in range(1,max_degree+1):
            factorial[0,i:]*=i
        orthonormalizer = 1 / np.sqrt(factorial)
        if derivative>0:
            y /= orthonormalizer
            for _ in range(derivative):
                y = np.concatenate([np.zeros([X.shape[0],1]),np.arange(1,y.shape[1])**1*y[:,:-1]],axis=1)
            y = 1/interval[1]**derivative*y
            y *= orthonormalizer
        return y
    elif measure == 't':
        y = taylor_polynomials(X,max_degree+1)
        if derivative>0:
            for _ in range(derivative):
                y = np.concatenate([np.zeros([X.shape[0],1]),np.arange(1,y.shape[1])*y[:,:-1]],axis=1)
        return y
    
def taylor_polynomials(X,N):
    return X.reshape(-1,1)**np.arange(N)

def chebyshev_polynomials(X, N):
    r'''
    Evaluate the orthonormal Chebyshev polynomials on
    :math:`([-1,1],dx/2)` in :math:`X\subset [-1,1]`
    
    :param X: Locations of desired evaluations
    :type X:  One dimensional np.array
    :param N: Number of polynomials
    :rtype: numpy.array of shape :code:`X.shape[0]xN`
    '''
    out = np.zeros((X.shape[0], N))
    deg = N - 1
    orthonormalizer = np.concatenate((np.array([1]).reshape(1, 1), np.sqrt(2) * np.ones((1, deg))), axis=1)
    if deg < 1:
        out = np.ones((X.shape[0], 1))
    else:
        out[:, 0] = np.ones((X.shape[0],))      
        out[:, 1] = X
        for n in range(1, deg):
            out[:, n + 1] = 2 * X * out[:, n] - out[:, n - 1]
    return out * orthonormalizer

def legendre_polynomials(X, N):
    r'''
    Evaluate the orthonormal Legendre polynomials on 
    :math:`([-1,1],dx/2)` in :math:`X\subset [-1,1]`
    
    :param X: Locations of desired evaluations
    :type X:  One dimensional np.array
    :param N: Number of polynomials
    :rtype: numpy.array of shape :code:`X.shape[0]xN`
    '''
    out = np.zeros((X.shape[0], N))
    deg = N - 1
    orthonormalizer = np.sqrt(2 * (np.arange(deg + 1)) + 1)
    if deg < 1:
        out = np.ones((X.shape[0], 1))
    else:
        out[:, 0] = np.ones((X.shape[0],))      
        out[:, 1] = X
        for n in range(1, deg):
            out[:, n + 1] = 1. / (n + 1) * ((2 * n + 1) * X * out[:, n] - n * out[:, n - 1])
    return out * orthonormalizer

def hermite_polynomials(X, N):
    r'''
    Evaluate the orthonormal Hermite polynomials on 
    :math:`(\mathbb{R},\frac{1}{\sqrt{2\pi}}\exp(-x^2/2)dx)` in :math:`X\subset\mathbb{R}`
    
    
    :param X: Locations of desired evaluations
    :type X:  One dimensional np.array
    :param N: Number of polynomials
    :rtype: numpy.array of shape :code:`X.shape[0] x N`
    '''
    out = np.zeros((X.shape[0], N))
    deg = N - 1
    factorial = np.ones((1,N))
    for i in range(1,N):
        factorial[0,i:]*=i
    orthonormalizer = 1 / np.sqrt(factorial)
    if deg < 1:
        out = np.ones((X.shape[0], 1))
    else:
        out[:, 0] = np.ones((X.shape[0],))      
        out[:, 1] = X
        for n in range(1, deg):
            out[:, n + 1] = X * out[:, n] - n * out[:, n - 1]
    return out * orthonormalizer



'''
Sparse approximation using Smolyak's algorithm

Usage:
    1) Setup a function instance that computes elements in a multi-index
    decomposition (and possibly auxiliary information about work and 
    contribution of the computed terms).
    Here, it can be helpful to use MixedDifferences from the module indices
    which turns regular algorithms into the form required for sparse
    approximations. As a special case, function approximation using 
    least squares polynomial approximations is already implemented in
    PolynomialApproximator.
    
    n = 2
    f = lambda x: np.sin(np.prod(x,1))
    pa = PolynomialApproximator(f,domain = [[0,1]]**n)

    2) Pass that function to a SparseApproximator instance, along
    with information about work and runtime estimates, etc.
    In the case of PolynomialApproximator instances, additional information
    is optional.

    sa = SparseApproximator(pa)

    3) Use :code:`update_approximation` to perform the computations

    sa.update_approximation(T=10)
    pa.get_approximation().plot()

'''
import copy
import warnings
import math
from timeit import default_timer as timer
import itertools
import random
import collections

import numpy as np
from numpy import Inf
from swutil.logs import Log
from swutil.decorators import log_calls
from swutil.collections import DefaultDict
from swutil.validation import NotPassed, Positive, Integer, Float, validate_args, \
    Nonnegative, Instance, DefaultGenerator, Function, Iterable, Bool, Dict, \
    List, Equals, InInterval, Arg, Passed, In
from swutil import plots

from smolyak.indices import  MultiIndexDict, get_admissible_indices, MISet, \
    kronecker, MultiIndex, get_bundles, get_bundle
from smolyak import indices
from smolyak.applications.polynomials import PolynomialApproximator

class _Factor:
    @validate_args('multipliers>(~func,~dims) multipliers==n',warnings=False)
    def __init__(self,
                 func:Function,
                 multipliers:Dict(value_spec=Positive & Float),
                 n:Positive & Integer,
                 dims:Function(value_spec=Bool)=lambda dim: True,
                 bundled:Bool=False,
                 ):
        self.bundled = bundled
        if Passed(multipliers):
            self.multipliers = multipliers
            func = lambda mi: np.prod([
                    self.multipliers[dim] ** mi[dim]
                    for dim in self.multipliers
            ])
            if self.bundled:
                self.func= lambda mis: sum(func(mi) for mi in mis)
            else:
                self.func = func
            self.have_none = False
            if math.isinf(n):
                self.have_all = False
            else:
                self.have_all = all(dim in self.multipliers for dim in range(n))
            self.dims = self.multipliers.__contains__
        else:
            self.multipliers = {}
            self.func = func
            self.dims = dims 
    def __call__(self, *args):
        return self.func(*args)
    
class WorkFunction(_Factor):
    pass

class ContributionFunction(_Factor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.bundled:
            raise ValueError('Contribution function cannot be bundled')

class _Decomposition:
    @validate_args(
        '~work_multipliers|~work_function',
        '~contribution_multipliers|~contribution_function',
        'bundled_dims>bundled',
        Arg('init_dims|next_dims', Passed) > Arg('n', Equals(math.inf)),
        warnings=False,
    )
    def __init__(
        self,
        Delta,
        n: Positive & Integer,
        work_multipliers: Dict(value_spec = InInterval(l = 1, lo = False), lenience = 2),
        work_function: Instance(WorkFunction),
        contribution_multipliers: Dict(value_spec = InInterval(l = 0, lo = True, r = 1, ro = True), lenience = 2),
        contribution_function: Instance(ContributionFunction),
        returns_work: Bool = False,
        returns_contributions: Bool = False,
        init_dims: List(Nonnegative & Integer, lenience = 2) = NotPassed,
        next_dims: Function = NotPassed,
        bundled: Bool = False,
        bundled_dims: Function(value_spec = Bool) | List(value_spec = Integer) = NotPassed,
        kronecker_exponents: Function(value_spec = Nonnegative & Float) = NotPassed,
        stores_approximation: Bool = False,
        structure=None,
        callback=None,
    ):
        r'''        
        :param Delta: Computes decomposition terms.
            In the most basic form, a single multi-index is passed to 
            :code:`Delta` and a single value, which represents the corresponding 
            decomposition term and supports vector space operations, is expected. 
            
            If :code:`returns_work == True`, :code:`Delta` must return (value,work) (or only work if stores_approximation),
            where work is a single float representing some abstract type of work
            that was required for the computation of value. Otherwise, algorithms
            are guided by runtime.
            
            If :code:`returns_contributions == True`, return (value,contribution) (or only 
            contribution if stores_approximation), where contribution is:
                *a single float if not stores_approximation and not bundled_dims or else
                *a dictionary  {(mi,contribution(float) for mi in bundle}
                    where bundle is the multi-index set that has been passed
            Otherwise, the return values will be normed by means of np.linalg.norm to
            assess their contribution, or they may implement norm() methods. 
            
            If both :code:`returns_contributions == True` and :code:`returns_work`
            then :code:`Delta` must return (value,work,contribution) (or only (work,contribution) in case of external storage)
            
            If :code:`bundled == True`, then :code:`Delta` is passed an iterable of multi-indices
            (which agree in the dimensions for which :code:`bundled_dims` returns True) must return 
            the sum of the corresponding decomposition elements.
            This may be useful for parallelization, or for problems where joint
            computation of decomposition elements is analytically more efficient.
            
            If :code:`stores_approximation == True`, no n_results is required and decomposition terms
            are expected to be stored by :code:`Delta` itself. Each call 
            will contain the full multi-index set representing the current approximation.
            
        :type Delta:  Function.
        :param n: Number of discretization parameters
        :type n: Integer, including math.inf (or 'inf') 
        :param work_multipliers: Specify factor by which work increases if index is increased in a given dimension
        :type work_multipliers: Dict
        :param work_function: If work is more complex, use work_function instead of work_multipliers to compute
            expected work associated with a given multi-index
        :type work_function: Function MultiIndex->Positive reals
        :param contribution_multipliers: Specify factor by which contribution decreases if index is increased in a given dimension
        :type contribution_multipliers: Dict
        :param contribution_function: If contribution is more complex, use contribution_function instead of contribution_multipliers to compute
            expected contribution of a given multi-index
        :type contribution_function: Function MultiIndex->Positive reals
        :param returns_work: Are the deltas returned together with their own work specification?
        :type returns_work: Boolean
        :param returns_contributions: Are the deltas returned together with their own contribution specification?
        :type returns_contributions: Boolean  
        :param init_dims: Initial dimensions used to create multi-index set. Defaults to :code:`range(n)`. However,
            for large or infinite `n`, it may make sense (or be necessary) to restrict this initially. 
            Each time a multi-index with non-zero entry in one of these initial dimensions, say j,  is selected, 
            new dimensions are added, according to output of :code:`next_dims(j)`
        :type init_dims: Integer or list of integers
        :param next_dims: Assigns to each dimension a list of child dimensions (see above)
        :type next_dims: :math:`\mathbb{N}\to 2^{\mathbb{N}}`
        :param: bundled: To be used when the decomposition terms cannot be computed independent of 
           each other. In this case, :code:`Delta` is called with subsets of :math:`\mathcal{I}` whose entries agree
           except for the dimensions specified in bundled_dims
        :type bundled: Boolean
        :param bundled_dims: Specifies dimensions used for bundling
        :type bundled_dims: :math:`\mathbb{N}\to\{\text{True},\text{False}\}` or list of dimensions
        :param kronecker_exponents: For infinite dimensional problems, the 
            contribution of Kronecker multi-index e_j is estimated as exp(kronecker_exponents(j))
        :type kronecker_exponents: Function from integers to negative reals  
        :param stores_approximation: Specifies whether approximation is computed externally. If True, :code:`Delta` is called multiple times 
            with a SparseIndex as argument (or a list of SparseIndices if :code:`bundled_dims` is True as well) and must collect the associated
            decomposition terms itself. If False, :code:`Delta` is also called multiple times with a SparseIndex (or a list of SparseIndices)
            and must return the associated decomposition terms, which are then stored within the SparseApproximator instance
        :type stores_approximation: Boolean
        :param structure: Used to enforce symmetries in approximating set.  
        :type structure: Function: multiindices->setsofmultiindices
        :param callback: If this returns a Trueish value, approximation is stopped (use for accuracy based approximations)
        :type callback: Function(): Bool
        '''
        if isinstance(Delta,PolynomialApproximator):
            if Passed(n) or returns_work:
                raise ValueError('Do not specify `n` or `returns_work` to SparseApproximator for polynomial approximation')  
            self.Delta = Delta.update_approximation
            self.n = Delta.n_acc+Delta.n
            _work_function = WorkFunction(func = Delta.estimated_work, dims = lambda n: (True if Passed(work_function) else n>=Delta.n_acc),bundled=True)
            self.returns_work = True 
            self.returns_contributions = True
            self.stores_approximation = True
            self.kronecker_exponents = kronecker_exponents 
            self._set_is_bundled(True,Delta.bundled_dims)
            self._set_work(work_multipliers,_work_function)
            self._set_contribution(contribution_multipliers,contribution_function)
            if isinstance(structure,str):
                if structure.lower() == 'td':
                    structure = lambda mi: [mi.restrict(lambda n:n<Delta.n_acc) + mi2.shifted(Delta.n_acc) 
                        for mi2 in indices.simplex(n = Delta.n,L=mi.mod(lambda n:n<Delta.n_acc).sum())]
                elif structure.lower() == 'pd':
                    structure = lambda mi: [mi.restrict(lambda n:n<Delta.n_acc) + mi2.shifted(Delta.n_acc) 
                        for mi2 in indices.rectangle(n = Delta.n,L=max(mi.mod(lambda n:n<Delta.n_acc)))]
                elif structure.lower() == 'sym':
                    from sympy.utilities.iterables import multiset_permutations
                    def structure(mi):
                        mim = mi.mod(lambda n:n<Delta.n_acc)
                        if mim==MultiIndex():
                            return []
                        else:
                            ret =  [mi.restrict(lambda n:n<Delta.n_acc) + mi2.shifted(Delta.n_acc)
                                for mi2 in [MultiIndex(perm) for perm in multiset_permutations(mi.mod(lambda n:n<Delta.n_acc).full_tuple())]]
                            return ret
        else:
            self.Delta = Delta
            self.n = n
            self.returns_work = returns_work
            self.returns_contributions = returns_contributions
            self.stores_approximation = stores_approximation
            self.kronecker_exponents = kronecker_exponents
            self._set_is_bundled(bundled, bundled_dims)
            self._set_work(work_multipliers, work_function)
            self._set_contribution(contribution_multipliers, contribution_function)
        self.callback = callback or (lambda: False)
        self.structure = structure or (lambda mi: set())
        if math.isinf(self.n):
            if not init_dims:
                init_dims = [0]
            self.init_dims = init_dims
            self._set_next_dims(next_dims)  
        else:
            self.init_dims = list(range(self.n))
        
    def _set_work(self, work_multipliers, work_function):
        if work_multipliers:
            self.work_function = WorkFunction(multipliers=work_multipliers, n=self.n,bundled=self.bundled)
        elif work_function:
            if work_function.bundled==self.bundled:
                self.work_function = work_function
            elif work_function.bundled and not self.bundled:
                raise ValueError("Work function cannot act on bundles when computations don't")
            elif self.bundled and not work_function.bundled:
                self.work_function = WorkFunction(func = lambda mis: sum(work_function(mi) for mi in mis),dims = work_function.dims,bundled=True)
        else:
            self.work_function = WorkFunction(func=lambda _: 1, dims=lambda dim: False, bundled=self.bundled)
        if self.bundled!=self.work_function.bundled:
            raise ValueError('If computations are bundled, work function must act on bundles as well (and conversely)')
        
    def _set_contribution(self, contribution_multipliers, contribution_function):
        if contribution_multipliers:
            self.contribution_function = ContributionFunction(multipliers=contribution_multipliers, n=self.n,bundled = False)
        elif contribution_function:
            self.contribution_function = contribution_function
        else:
            self.contribution_function = ContributionFunction(func=lambda _: 1, dims=lambda dim:False, bundled=False)

    def _set_is_bundled(self, bundled, bundled_dims):
        self.bundled = bundled
        if hasattr(bundled_dims, '__contains__'):
            self.bundled_dims = lambda dim: dim in bundled_dims
        else: 
            self.bundled_dims = bundled_dims
         
    def _set_next_dims(self, next_dims):
        if (not next_dims) and self.n == Inf:
            self.next_dims = lambda dim: [dim + 1] if dim + 1 not in self.init_dims else []
        else:
            self.next_dims = next_dims

class SparseApproximator:
    r'''
    Sparse approximation based on multi-index decomposition.

    Given a decomposition of :math:`f_{\infty}` as    
    .. math::

        f_{\infty}=\sum_{\mathbf{k}\in\mathbb{N}^{n}} (\Delta f)(\mathbf{k}),

    this class computes and stores approximations of the form
    .. math:: 

       \mathcal{S}_{\mathcal{I}}f:=\sum_{\mathbf{k}\in\mathcal{I}} (\Delta f)(\mathbf{k}),

    for finite index sets :math:`\mathcal{I}\subset\mathbb{R}^n`. 

    To compute the approximation, use `update_approximation`. This method provides
    multiple ways to choose the index set :math:`\mathcal{I}`. 
    '''
    @validate_args(
        '~work_multipliers|~work_function',
        '~contribution_multipliers|~contribution_function',
        'bundled_dims>bundled',
        Arg('init_dims|next_dims', Passed) > Arg('n', Equals(math.inf)),
        warnings=False,
    )
    def __init__(
        self,
        Delta,
        n: Positive & Integer,
        work_multipliers: Dict(value_spec = InInterval(l = 1, lo = False), lenience = 2),
        work_function: Instance(WorkFunction),
        contribution_multipliers: Dict(value_spec = InInterval(l = 0, lo = True, r = 1, ro = True), lenience = 2),
        contribution_function: Instance(ContributionFunction),
        returns_work: Bool = False,
        returns_contributions: Bool = False,
        init_dims: List(Nonnegative & Integer, lenience = 2) = NotPassed,
        next_dims: Function = NotPassed,
        bundled: Bool = False,
        bundled_dims: Function(value_spec = Bool) | List(value_spec = Integer) = NotPassed,
        kronecker_exponents: Function(value_spec = Nonnegative & Float) = NotPassed,
        stores_approximation: Bool = False,
        structure=None,
        callback=None,
    ):
        r'''        
        :param Delta: Computes decomposition terms.
            In the most basic form, a single multi-index is passed to 
            :code:`Delta` and a single value, which represents the corresponding 
            decomposition term and supports vector space operations, is expected. 
            
            If :code:`returns_work == True`, :code:`Delta` must return (value,work) (or only work if stores_approximation),
            where work is a single float representing some abstract type of work
            that was required for the computation of value. Otherwise, algorithms
            are guided by runtime.
            
            If :code:`returns_contributions == True`, return (value,contribution) (or only 
            contribution if stores_approximation), where contribution is:
                *a single float if not stores_approximation and not bundled_dims or else
                *a dictionary  {(mi,contribution(float) for mi in bundle}
                    where bundle is the multi-index set that has been passed
            Otherwise, the return values will be normed by means of np.linalg.norm to
            assess their contribution, or they may implement norm() methods. 
            
            If both :code:`returns_contributions == True` and :code:`returns_work`
            then :code:`Delta` must return (value,work,contribution) (or only (work,contribution) in case of external storage)
            
            If :code:`bundled == True`, then :code:`Delta` is passed an iterable of multi-indices
            (which agree in the dimensions for which :code:`bundled_dims` returns True) must return 
            the sum of the corresponding decomposition elements.
            This may be useful for parallelization, or for problems where joint
            computation of decomposition elements is analytically more efficient.
            
            If :code:`stores_approximation == True`, no n_results is required and decomposition terms
            are expected to be stored by :code:`Delta` itself. Each call 
            will contain the full multi-index set representing the current approximation.
            
        :type Delta:  Function.
        :param n: Number of discretization parameters
        :type n: Integer, including math.inf (or 'inf') 
        :param work_multipliers: Specify factor by which work increases if index is increased in a given dimension
        :type work_multipliers: Dict
        :param work_function: If work is more complex, use work_function instead of work_multipliers to compute
            expected work associated with a given multi-index
        :type work_function: Function MultiIndex->Positive reals
        :param contribution_multipliers: Specify factor by which contribution decreases if index is increased in a given dimension
        :type work_multipliers: Dict
        :param work_function: If contribution is more complex, use contribution_function instead of contribution_multipliers to compute
            expected contribution of a given multi-index
        :type work_function: Function MultiIndex->Positive reals
        :param returns_work: Does the decomposition come with its own cost specification?
        :type returns_work: Boolean
        :param returns_contributions: Does the decomposition come with its own contribution specification? 
        :type returns_contributions: Boolean  
        :param init_dims: Initial dimensions used to create multi-index set. Defaults to :code:`range(n)`. However,
            for large or infinite `n`, it may make sense (or be necessary) to restrict this initially. 
            Each time a multi-index with non-zero entry in one of these initial dimensions, say j,  is selected, 
            new dimensions are added, according to output of :code:`next_dims(j)`
        :type init_dims: Integer or list of integers
        :param next_dims: Assigns to each dimension a list of child dimensions (see above)
        :type next_dims: :math:`\mathbb{N}\to 2^{\mathbb{N}}`
        :param: bundled: To be used when the decomposition terms cannot be computed independent of 
           each other. In this case, :code:`Delta` is called with subsets of :math:`\mathcal{I}` whose entries agree
           except for the dimensions specified in bundled_dims
        :type bundled: Boolean
        :param bundled_dims: Specifies dimensions used for bundling
        :type bundled_dims: :math:`\mathbb{N}\to\{\text{True},\text{False}\}` or list of dimensions
        :param kronecker_exponents: For infinite dimensional problems, the 
            contribution of Kronecker multi-index e_j is estimated as exp(kronecker_exponents(j))
        :type kronecker_exponents: Function from integers to negative reals  
        :param stores_approximation: Specifies whether approximation is computed externally. If True, :code:`Delta` is called multiple times 
            with a SparseIndex as argument (or a list of SparseIndices if :code:`bundled_dims` is True as well) and must collect the associated
            decomposition terms itself. If False, :code:`Delta` is also called multiple times with a SparseIndex (or a list of SparseIndices)
            and must return the associated decomposition terms, which are then stored within the SparseApproximator instance
        :type stores_approximation: Boolean
        :param structure: Used to enforce symmetries in approximating set.  
        :type structure: Function: multiindices->setsofmultiindices
        :param callback: If this returns a Trueish value, approximation is stopped (use for accuracy based approximations)
        :type callback: Function(): Bool
        '''
        self.decomposition = _Decomposition(
            Delta,
            n,
            work_multipliers,
            work_function,
            contribution_multipliers,
            contribution_function,
            returns_work,
            returns_contributions,
            init_dims,
            next_dims,
            bundled,
            bundled_dims,
            kronecker_exponents,
            stores_approximation,
            structure,
            callback,
        )
        self.log = Log(print_filter=False)
        self.data = _Data(self.decomposition)
        
    def update_approximation(self, mode=None,indices=None, T=None, L=None):
        '''
        Update the multi-index set and compute the resulting updated approximation.

        There are four ways to determine the updated multi-index set:
         :code:`indices`, which requires passing the new multi-indices with the argument `indices`
         :code:`adaptive`, which increases the multi-index set adaptively, one multi-index at a time
         :code:`apriori`, which constructs the set based on a-priori knowledge about work and contribution of the decomposition terms
         :code:`continuation`, which constructs the set by a combination of first learning the behavior of work and contribution, and then using this knowledge to create optimal sets.
        
        ADAPTIVE:
            To decide on the multi-index to be added at each step, estimates of contributions and work are maintained. 

            These estimates are based on neighbors that are already in the set :math:`\mathcal{I}`,
            unless they are specified in :code:`contribution_factors` and :code:`work_factors`.
            If user defines :code:`have_work_factor` and :code:`have_contribution_factor` 
            that only estimates for some of the :code:`n` involved parameters are available, 
            then the estimates from :code:`contribution_factor` and :code:`work_factor` for those parameters
            are combined with neighbor estimates for the remaining parameters.
        
            Must pass exactly one of the following additional arguments:
            :param N: Maximal number of new multi-indices.
            :type N: Integer
            :param T: Maximal time (in seconds).
            :type T: Positive

        APRIORI:
            Use apriori estimates of contributions and work provided to determine.
        
            Must pass exactly one of the following additional arguments:
            :param L: Threshold 
            :type L: Positive
            :param T: Maximal time (in seconds).
            :type T: Positive

        CONTINUATION:
            In an iterative manner, determine optimal multi-index-simplices by fitting contribution
            and work parameters to simplex of previous iteration.

            :param T: Maximal time (in seconds).
            :type T: Positive

        INDICES:
            Use user-specified multi-index set 
            
            :param indices: Multi-index set
            :type indices: iterable of MultiIndex instances

        :param mode: Update mode
        :type mode: One of 'indices', 'adaptive', 'apriori', 'continuation'
        '''
        tic_init = timer()
        Passed = lambda x: x is not None
        if not Passed(mode):
            if Passed(indices): 
                mode = 'indices'
            if Passed(T):
                mode = 'adaptive'
            if Passed(L):
                mode = 'apriori'
        if mode == 'indices':
            if Passed(L) or Passed(T):
                raise ValueError('Cannot pass L or T in indices mode')
            it = [0]
        elif mode == 'apriori':
            if Passed(T) == Passed(L):
                raise ValueError('Must pass either L or T in apriori mode')
            if Passed(L):
                it = [L]
            else:
                it = itertools.count()
        elif mode == 'adaptive':
            if Passed(T) == Passed(L):
                raise ValueError('Must pass either L or T in adaptive mode')
            if self.decomposition.bundled and not self.decomposition.returns_contributions:
                raise ValueError('Cannot run adaptively when decomposition.bundled but not decomposition.returns_contributions')
            if Passed(L):
                it = range(L)
            else:
                it = itertools.count()
        elif mode== 'continuation':
            if Passed('L'):
                raise ValueError('Cannot pass L in continuation mode')
            if np.isinf(self.decomposition.n):
                raise ValueError('Cannot use continuation for infinite-dimensional problems')
            it = itertools.count()
            n = self.decomposition.n
            work_exponents = lambda: [
                self._get_work_exponent(i) 
                if (
                        self.data.work_model_estimator if 
                        self.decomposition.returns_work 
                        else self.data.runtime_estimator
                ).ratios[i] 
                else 1 
                for i in range(n)
            ]
            contribution_exponents = lambda: [
                self._get_contribution_exponent(i) 
                if 
                self.data.contribution_estimator.ratios[i]
                else 1 
                for i in range(n)
            ]
        else: 
            raise ValueError('No mode selected')
        max_time=0
        for l in it:
            tic_iter = timer()
            if Passed(T) and l>0 and max_time + (timer() - tic_init) > T: 
                break
            if self.decomposition.callback():
                break
            if mode == 'apriori':
                mis_update = self.data.apriori_indices(l)
            elif mode == 'adaptive':
                mis_update = [self.data.next_best_mi()]
            elif mode == 'continuation':
                scale = min(work_exponents() + contribution_exponents() for dim in range(n))
                mis_update = indices.simplex(L=l, weights=(work_exponents() + contribution_exponents()) / scale, n=n)
            elif mode == 'indices':
                mis_update = indices
            self._extend(mis_update) # check whether this rbreaks when mis_udpate is a subset of current mis
            max_time = max(timer() - tic_iter,max_time)
        
    @log_calls    
    def _extend(self, mis):
        '''
        :param mis: Multi-indices to add to approximation
        :type mis: Iterable of multi-indices
        '''
        if self.decomposition.bundled:
            miss = get_bundles(mis, self.decomposition.bundled_dims)
            not_bundled_dims = lambda dim: not self.decomposition.bundled_dims(dim)
            key = lambda mis: mis[0].restrict(not_bundled_dims)
            it = sorted(miss, key=key)
        else:
            it = mis
        for temp in it:
            work_model = None
            contribution = None
            object_slice = None
            if self.decomposition.bundled:
                mi_update = temp[0]
                mis_update = temp
            else:
                mi_update = temp
                mis_update = [temp]
            if not self.decomposition.stores_approximation and not self.decomposition.Delta:  # Dry run
                return  
            if self.decomposition.bundled:
                external_work_factor = self.decomposition.work_function(mis_update) # Want to be able to provide work function with whole bundle without checking what is actually new and then asking each of those separately, thats why it is required that work_function covers all bundled dimensions
            else:
                external_work_factor = self.decomposition.work_function(mi_update)
            self.data.extend(mis_update)
            if self.decomposition.stores_approximation:
                argument = self.data.mis.mis # provide full set, leave it to external to reuse computations or not 
            else:
                argument = mis_update if self.decomposition.bundled else mi_update
            tic = timer()
            output = self.decomposition.Delta(copy.deepcopy(argument))
            runtime = timer() - tic
            n_arg = sum(map(int, [not self.decomposition.stores_approximation, self.decomposition.returns_work, self.decomposition.returns_contributions]))  # Possible outputs: Decomposition term, work, contribution
            if n_arg == 1 and not isinstance(output, tuple):  # Allow user to not return tuples if not necessary
                output = [output]
            if not self.decomposition.stores_approximation:  # Decomposition term
                object_slice, output = output[0], output[1:]
                output = output[1:]
            if self.decomposition.returns_work and self.decomposition.returns_contributions: # Remaining 2 outputs
                work_model, contribution = output
            elif self.decomposition.returns_work and not self.decomposition.returns_contributions: # Remaining 1 output
                work_model = output[0]
            elif not self.decomposition.returns_work and self.decomposition.returns_contributions:  # Remaining 1 output
                contribution = output[0]
            if self.decomposition.returns_contributions:  # User decides what contribution means
                if not self.decomposition.bundled_dims and not Dict.valid(contribution):  # Allow user to not return dictionary if not_bundled (which means user is only passed single multi-index)
                    contribution = {mi_update: contribution}
            elif not self.decomposition.stores_approximation:  # If approximation is stored here instead of by user, try to figure out contribution
                try:
                    if self.decomposition.bundled_dims:  # If approximation is grouped into bundles, norm function must be able to divide contribution into single multi-indices
                        contribution = {mi: object_slice.norm(mi) for mi in mis_update}
                    else:
                        contribution = {mi_update: object_slice.norm()}
                except AttributeError:  # User didn't implement .norm()
                    try:
                        if self.decomposition.bundled_dims:  # TODO: Does it make sense to assign each term in bundle the same contribution? 
                            contribution = {mi: np.linalg.norm(object_slice) for mi in mis_update}
                        else:  # Only one new term in approximation
                            contribution = {mi_update: np.linalg.norm(object_slice)}
                    except AttributeError:
                        pass
            if self.decomposition.returns_contributions:
                if contribution is None:
                    raise ValueError("Decomposition didn't return contributions") 
                if set(contribution.keys())!=set(argument):
                    raise ValueError('Contributions did not match multi-index set')
            if self.decomposition.returns_work and work_model == None:
                raise ValueError("Decomposition didn't return work")
            self.data.update_estimates(mis_update, mi_update, object_slice, contribution, work_model, runtime, external_work_factor)
             
    def get_approximation(self):
        if self.decomposition.stores_approximation:
            raise ValueError('Decomposition is stored externally')
        else:
            return sum([self.data.object_slices[mi] for mi in self.data.object_slices])   
    
    def get_contribution_multiplier(self, dim):
        return np.exp(-self._get_contribution_exponent(dim))

    def get_runtime_multiplier(self, dim):
        if self.decomposition.returns_work:
            raise ValueError('Since decomposition provides abstract work model, no runtime estimates are kept. Try `get_work_multiplier` instead.')
        return np.exp(self._get_runtime_exponent(dim))

    def get_work_multiplier(self, dim):
        return np.exp(self._get_work_exponent(dim))
    
    def get_total_work_model(self):
        return sum(self.data.work_models._dict.values())
    
    def get_total_runtime(self):
        return sum(self.data.runtimes._dict.values())
    
    def get_indices(self):
        return copy.deepcopy(self.data.mis.mis)

    def _get_work_exponent(self, dim):
        estimator = self.data.work_model_estimator if self.decomposition.returns_work else self.data.runtime_estimator
        if not estimator.dims_ignore(dim):
                return estimator.exponents[dim]
        else:
            raise KeyError('No work fit for this dimension')
    
    def _get_runtime_exponent(self, dim):
        if not self.data.runtime_estimator.dims_ignore(dim):
            return self.data.runtime_estimator.exponents[dim]
        else:
            raise KeyError('No runtime fit for this dimension')
       
    def _get_contribution_exponent(self, dim):
        if not self.data.contribution_estimator.dims_ignore(dim):
            return -self.data.contribution_estimator.exponents[dim]
        else:
            raise KeyError('No contribution fit for this dimension') 
    
    @validate_args(warnings=False)
    def plot_indices(self, dims:Iterable, weights:In('contribution','work_model','runtime','contribution/work_model','contribution/runtime',False)=False, percentiles:Positive & Integer=4):
        '''
        :param dims: Dimensions that should be used for plotting
        :type dims: List of integers, length at most 3
        :param weights: Determines size of points
        :type weights: 'contribution' or 'work_model' or 'runtime' or 'contribution/work_model' or 'contribution/runtime'
        :param percentiles: Plot given number of weight-percentile groups in different colors
        :type perentiles: Integer
        '''
        if NotPassed(dims):
            dims = list(self.data.mis.active_dims)
        if not weights:
            percentiles = 1
            weight_dict = None
        elif weights == 'contribution':
            weight_dict = {mi: self.data.contributions[mi] for mi in self.get_indices()}
        elif weights == 'work_model':
            if not self.decomposition.returns_work:
                raise ValueError('Decomposition does not provide abstract work model')
            weight_dict = {mi: self.data.work_models[mi] for mi in self.get_indices()}
        elif weights == 'runtime':
            weight_dict = {mi: self.data.runtimes[mi] for mi in self.get_indices()}
        elif weights == 'contribution/work_model':
            assert(self.decomposition.returns_work)
            weight_dict = {mi:self.data.contributions[mi] / self.data.work_models[mi] for mi in self.get_indices()}
        elif weights == 'contribution/runtime':
            weight_dict = {mi: self.data.contributions[mi] / self.data.runtimes[mi] for mi in self.get_indices()}
        else: 
            raise ValueError('Cannot use weights {}'.format(weights))
        plots.plot_indices(mis=self.get_indices(), dims=dims, weights=weight_dict, groups=percentiles) 

class _Estimator:
    def __init__(self, dims_ignore, exponent_max, exponent_min, init_exponents=None,name=''):
        self.quantities = {}
        self.dims_ignore = dims_ignore
        self.FIT_WINDOW = int(1e6)
        self.ratios = DefaultDict(lambda dim: collections.deque([],self.FIT_WINDOW))
        init_exponents = init_exponents or (lambda dim:0)
        self.fallback_exponents = DefaultDict(init_exponents)  # USED AS PRIOR IN EXPONENT ESTIMATION AND AS INITIAL GUESS OF EXPONENT WHEN NO DATA AVAILABLE AT ALL
        self.exponents = DefaultDict(lambda dim: self.fallback_exponents[dim])
        self.exponent_max = exponent_max
        self.exponent_min = exponent_min
        self.active_dims = set()   
        self.name=name
        
    def set_fallback_exponent(self, dim, fallback_exponent):
        self.fallback_exponents[dim] = fallback_exponent
        
    def __contains__(self, mi):
        mi = mi.mod(self.dims_ignore)
        return mi in self.quantities
    
    def __setitem__(self, mi, q):
        self.active_dims.update(set(mi.active_dims()))
        mi = mi.mod(self.dims_ignore)
        q = float(q)
        have = mi in self.quantities
        self.quantities[mi] = q  #two reasons to overwrite: 1) in least squares polynomials for the contribution estimate, the estimate of every single coefficient gets better and better over time 2) in general, for dimensions that are modulod out because their work contribution factor is known, this entry is theoretically the same but practically different. for example, look at MLMC, assume the work factor of the first parameter is theoretically 2. then this stores effectively the cost per sample divided by 2**l. however, this cost is not actually indpendent of the level  
        if not have:
            get_ratio = lambda a,b: a/b if b>0 else (np.Inf if a>0 else 1)
            for dim in self.active_dims:
                neighbor = mi + (-1)*kronecker(dim)
                if neighbor in self.quantities:
                    self.ratios[dim].append(get_ratio(q,self.quantities[neighbor])) 
            self._update_exponents()
        
    def _update_exponents(self):
        for dim in self.ratios:
            ratios = np.array(self.ratios[dim])
            estimate = max(min(np.median(ratios), np.exp(self.exponent_max)), np.exp(self.exponent_min))
            c = len(ratios)
            self.exponents[dim] = (self.fallback_exponents[dim] + c * np.log(estimate)) / (c + 1.)
    
    def __call__(self, mi):
        mi = mi.mod(self.dims_ignore)
        if mi in self.quantities:
            return self.quantities[mi]
        else:
            if mi.is_kronecker():
                dim = mi.active_dims()[0]
                return self.quantities[MultiIndex()]*np.exp(self.exponents[dim])
            else:
                estimate = 0
                for dim,sign in itertools.product(self.active_dims,(-1,1)):
                    neighbor = mi + sign*kronecker(dim)
                    if neighbor in self.quantities:
                        estimate = max(estimate,self.quantities[neighbor])
                return estimate

class _Data:
    def __init__(self, decomposition):
        self.WORK_EXPONENT_MAX = 10
        self.WORK_EXPONENT_MIN = 0
        self.CONTRIBUTION_EXPONENT_MAX = 0
        self.CONTRIBUTION_EXPONENT_MIN = -10
        self.decomposition = decomposition 
        # If decomposition is bundled, runtime_estimator only makes sense if runtime_function has all bundled dimensions. runtime_function may or not be bundled; if it is not it will be called for each index in bundled and the observed runtime will then be divided by the sum 
        # work_model function does not have to be bundled if decomposition is. if it is bundled, it does have to include bundled dims in its dims
        # contribution function cannot be bundled
        self.runtime_estimator = _Estimator(self.decomposition.work_function.dims,
                                      exponent_max=self.WORK_EXPONENT_MAX,
                                      exponent_min=self.WORK_EXPONENT_MIN,
                                      name='runtime')
        self.runtimes = MultiIndexDict(lambda dim: self.decomposition.bundled and self.decomposition.bundled_dims(dim))
        self.contribution_estimator = _Estimator(self.decomposition.contribution_function.dims,
                                              exponent_max=self.CONTRIBUTION_EXPONENT_MAX,
                                              exponent_min=self.CONTRIBUTION_EXPONENT_MIN,
                                              init_exponents=self.decomposition.kronecker_exponents,
                                              name='contribution')
        if self.decomposition.returns_work:
            self.work_model_estimator = _Estimator(self.decomposition.work_function.dims,
                                      exponent_max=self.WORK_EXPONENT_MAX,
                                      exponent_min=self.WORK_EXPONENT_MIN,
                                      name='work_model')
            self.work_models = MultiIndexDict(decomposition.bundled_dims)
        self.mis = MISet(dims=decomposition.init_dims)
        self.mis.structure_constraints = set([MultiIndex()])
        self.mis.structure_constraints |= set(MultiIndex(((i,1),),sparse=True) for i in self.decomposition.init_dims)
        self.object_slices = MultiIndexDict(decomposition.bundled_dims)
        self.contributions = dict()

    def next_best_mi(self):
        mis = self.mis
        for mi in mis.structure_constraints.copy():
            if mi in mis:
                mis.structure_constraints.discard(mi)
            elif mis.is_admissible(mi):
                mi_update = mi
                break
        else:
            estimates = {mi:self.profit_estimate(mi) for mi in mis.candidates} 
            # plots.plot_indices(mis.candidates)
            # plots.plot_indices(mis.mis,colors=[[0,0,0]])
            # import matplotlib.pyplot as plt
            # plt.show()
            mi_update = max(mis.candidates, key=lambda mi: self.profit_estimate(mi))
            if self.decomposition.structure:
                mis.structure_constraints |= set(self.decomposition.structure(mi_update))
        return mi_update

    def apriori_indices(self,L):
        def admissible(mi):
            return self.profit_estimate(mi) ** (-1) <= np.exp(L + 1e-12)
        try:
            mis = get_admissible_indices(lambda mi: admissible(mi), self.decomposition.n)
        except KeyError:
            raise KeyError('Did you specify the work for all dimensions?')
    
    def extend(self,mis_update):
        if math.isinf(self.decomposition.n):
            self._find_new_dims(mis_update)
        self.mis.update(mis_update)        
        
    def update_estimates(self, mis_update, mi_update, object_slice, contribution, work_model, runtime, external_work_factor):
        self.object_slices[mi_update] = object_slice
        self.runtimes[mi_update] = runtime
        if self.decomposition.returns_work:  # external_work_factor refers to work_model 
            if external_work_factor>0:
                self.work_model_estimator[mi_update] = work_model / external_work_factor  # Here lies reason why work factor must be bundled if decomposition is bundled: `work_model` does not distinguish different contributions to work (its for the entire mi_update-bundle), so external_work_factor must also make its predictions for entires bundles
            self.work_models[mi_update] = work_model    
        else: # external_work_factor refers to runtime
            self.runtime_estimator[mi_update] = runtime / external_work_factor
        try:
            for mi in contribution:
                self.contribution_estimator[mi] = contribution[mi] / self.decomposition.contribution_function(mi) # Here lies reason why the reasoning above does not apply to contributions: they can always be split up among all the multi-indices in a bundle (this is implicit in the assumption that the used provided contributions are separate for each mi in mi_update) 
                self.contributions[mi] = contribution[mi]
        except (KeyError, NameError): 
            pass  # Contribution could not be determined, contribution was never created 
    
    def profit_estimate(self, mi):
        contribution = self.contribution_estimator(mi) * self.decomposition.contribution_function(mi)
        if self.decomposition.bundled:
            mi_to_bundle = lambda mi: get_bundle(mi, self.mis, self.decomposition.bundled_dims) + [mi]
        else:
            mi_to_bundle = lambda mi: mi
        if self.decomposition.returns_work:
            work = self.work_model_estimator(mi) * self.decomposition.work_function(mi_to_bundle(mi))
            return contribution / work
        else:
            runtime = self.runtime_estimator(mi) * self.decomposition.work_function(mi_to_bundle(mi))
            return contribution / runtime
        
    def _find_new_dims(self, mis_update): 
        for mi in mis_update:
            if mi.is_kronecker() and not mi in self.mis:
                dim_trigger = mi.active_dims()[0]
                dims_new = self.decomposition.next_dims(dim_trigger)
                for dim in dims_new:
                    self.mis.add_dimensions([dim])
                    if self.decomposition.returns_work:
                        self.work_model_estimator.set_fallback_exponent(dim, self.work_model_estimator.exponents[dim_trigger])
                    else:
                        self.runtime_estimator.set_fallback_exponent(dim, self.runtime_estimator.exponents[dim_trigger])
                    if not self.decomposition.kronecker_exponents:
                        self.contribution_estimator.set_fallback_exponent(dim, self.contribution_estimator.exponents[dim_trigger])




import warnings
import math
import copy
import timeit
import functools

import numpy as np
import scipy
from scipy.linalg import solve
from scipy.linalg.misc import LinAlgError
import matplotlib.pyplot as plt
from matplotlib import cm
#from mpl_toolkits.mplot3d import Axes3D  # @UnusedImport @UnresolvedImport

from swutil.np_tools import grid_evaluation
from swutil.collections import RFunction, VFunction
from swutil.collections import DefaultDict
from swutil.decorators import print_runtime
from swutil.np_tools import grid_evaluation
from swutil.time import Timer
import swutil.validation

from smolyak.applications.polynomials import samples
from smolyak.indices import MixedDifferences
from smolyak import indices
from smolyak.indices import MultiIndex, cartesian_product
from smolyak.applications.polynomials.probability_distributions import ProbabilityDistribution,\
    ProductProbabilityDistribution, AbstractProbabilityDistribution 
from smolyak.applications.polynomials.orthogonal_polynomials import evaluate_orthonormal_polynomials

class PolynomialSpace:
    '''
    Vector space of polynomials for least-squares polynomial approximation on open 
    subsets of Euclidean space.
    
    Combines domain (in form of probability space) with description
    of finite-dimensional subspace of polynomials

    :param n: Dimension of domain
    :param k: Polynomial degree
    '''
    def __init__(self, probability_distribution = None,basis=None,warnings = False, n = None, k = None):
        if isinstance(probability_distribution,ProbabilityDistribution):
            self.probability_distribution = ProductProbabilityDistribution([probability_distribution])
        else:
            if probability_distribution is None:
                probability_distribution = 't'
            self.probability_distribution = ProbabilityDistribution(probability_distribution)**n
            if basis is None and k is None:
                raise ValueError('Must provide basis or polynomial degree')
            basis = basis or indices.simplex(L=k,n=n)
        self.set_basis(basis)
        self.warnings = warnings
    
    def get_dimension(self): 
        '''
        Return dimension of polynomial subspace (not of domain, use get_c_var() for that)
        '''
        return len(self.basis)

    @property
    def dimension(self):
        return self.get_dimension()
    
    def optimal_weights(self, X):
        '''
        Return inverse of optimal sample density (=normalized Christoffel function)
        '''
        if self.get_dimension()>0:
            if X.shape[0] == 0:
                return np.zeros((0, 1))
            else:
                B = self.evaluate_basis(X)
                W = self.get_dimension() / np.sum(np.power(B, 2), axis=1)
                return W   
        else:
            raise ValueError('Polynomial subspace is zero-dimensional')
    
    def get_approximation(self, X, Y, W=None):
        coeff,*_ = self.weighted_least_squares(X=X,Y=Y,W=W)
        return PolynomialApproximation(self,coeff)
        
    def weighted_least_squares(self, X, Y, W=None, basis_extension=None):
        '''
        Compute least-squares approximation. 
        
        :param X: sample locations
        :type X: `N x self.get_c_var()` numpy array
        :param W: weights
        :param Y: sample values
        :return: coefficients
        '''
        W = W if W is not None else np.ones([len(X)])
        B = self.evaluate_basis(X)
        if basis_extension:
            b_extra = basis_extension(X)
            B = np.concatenate([B,b_extra],axis=1)
        W = W.reshape(-1)
        Y = Y.reshape(-1)
        cond = None
        if self.warnings:
            G = B.transpose().dot(B * W)
            cond = np.linalg.cond(G)
            if cond > 3:
                warnings.warn('Ill conditioned Gramian matrix encountered') 
        if B.shape[1]>0:
            tol=1e-9
            rls = np.sqrt(W)*Y
            M = np.sqrt(W)[:,None]*B
            coefficients,*info = scipy.sparse.linalg.lsmr(M,rls,atol=tol,btol=tol)
            if self.warnings and not np.isfinite(coefficients).all():
                warnings.warn('Numerical instability encountered')
        if basis_extension:
            return {pol: coefficients[i] for i, pol in enumerate(self.basis)},cond, coefficients[len(self.basis):]
        else:
            return {pol: coefficients[i] for i, pol in enumerate(self.basis)},cond
         
    def get_active_dims(self):
        if self.basis:
            return set.union(*[set(pol.active_dims()) for pol in self.basis]) 
        else:
            return set()
    
    def get_c_var(self):
        '''
        Return dimension of domain (not of polynomial subspace, use get_dimension() for that)
        '''
        return self.probability_distribution.get_c_var()

    def evaluate_basis(self, X,derivative=None):
        '''
        Evaluates basis polynomials at given sample locations.
        
        :param X: Sample locations
        :type X: Numpy array of shape `N x self.get_c_var()` 
        :return: Basis polynomials evaluated at X
        :rtype: `N x self.get_dimension()` np.array
        '''
        values = np.ones((X.shape[0], self.get_dimension()))
        basis_values = {}
        if self.basis:
            active_vars = sorted(list(set.union(*[set(pol.active_dims()) for pol in self.basis])))
        else:
            active_vars = list()
        if not derivative:
            derivative = np.zeros(self.get_c_var())
        elif len(derivative)<self.get_c_var():
            derivative = derivative + [0]*(self.get_c_var()-len(derivative))
        if any(order>0 and dim not in active_vars for dim,order in enumerate(derivative)):
            values[:] = 0
        elif active_vars:
            for dim in active_vars:
                basis_values[dim] = evaluate_orthonormal_polynomials(
                    X[:,dim],
                    max(pol[dim] for pol in self.basis), 
                    measure=self.probability_distribution.ups[dim].measure, 
                    interval=self.probability_distribution.ups[dim].interval,
                    derivative = derivative[dim]
                )
            projected_basis = [pol.retract(lambda i: active_vars[i]).full_tuple(c_dim = len(active_vars)) for pol in self.basis]
            for i,pol in enumerate(self.basis):
                if any(order>0 and dim not in pol.active_dims() for dim,order in enumerate(derivative)):#Check necessity 
                    values[:,i] = 0                                                                                      
                else:
                    values[:,i] = np.prod([basis_values[active_vars[d]][:,projected_basis[i][d]] for d in range(len(active_vars))],axis=0)
        return values
    
    def set_basis(self,polynomials):
        self.basis=[]
        if polynomials:
            self.extend_basis(polynomials)
        
    def extend_basis(self, polynomials):
        '''
        Extend polynomial subspace.
        
        If necessary, extend dimension of domain as well. 
        
        :param polynomials: Polynomials to be added
        '''
        self.basis += polynomials
        if self.basis:
            max_var = max([pol.max_dim()for pol in self.basis])
            if max_var > self.get_c_var():
                for __ in range(max_var - self.get_c_var()):
                    self.probability_distribution.ups.append(copy.deepcopy(self.probability_distribution.ups[-1]))
        
    def plot_optimal_distribution(self, N=200, L=1):
        '''
        Plot optimal sampling distribution
        '''
        if self.get_c_var() == 1:
            X = self.probability_distribution.get_range(N, L)
            Z = self.probability_distribution.lebesgue_density(X) / self.optimal_weights(X)
            fig = plt.figure()
            ax = fig.gca()
            ax.plot(X, Z)
        elif self.get_c_var() == 2:
            X, Y = self.probability_distribution.get_range(N, L)
            Z = grid_evaluation(X, Y,self.probability_distribution.lebesgue_density) / grid_evaluation(X, Y,self.optimal_weights)
            fig = plt.figure()
            ax = fig.gca(projection='3d')
            ax.plot_surface(X, Y, Z, cmap=cm.coolwarm, alpha=0.5)  # @UndefinedVariable
        fig.suptitle('Lebesgue density of optimal distribution')
        plt.show()
        return np.sum(Z) / Z.size

class PolynomialApproximator:
    r'''
    Maintains polynomial approximation of a given function 
    :math:`f\colon [a,b]^d\to\mathbb{R}`
    that can be sampled at different levels of accuracy.
    '''
    def __init__(self, function=None,domain=None,C=1,sampler='optimal',reparametrization=False,n = None, n_acc=0,warnings = False):
        r'''
        :param function: Function :math:`[a,b]^d\to\mathbb{R}` that is being
        approximated. Needs to support :code:`__call__(X,mi)` where :code:`X` 
        is an np.array of size N x :math:`d` describing the sample locations and :code:`mi` is a
        multi-index describing the required accuracy of the samples.
        :param cdim_acc: Number of discretization parameters of `self.function`
        (=length of `mi` above)
        :param domain: Probability distribution
        :type domain: List of intervals or ProbabilityDistribution
        :param C: see SinglelevelPolynomialApproximator
        :param sampler: see SinglelevelPolynomialApproximator
        :param reparametrization: Determines whether polynomial subspaces are indexed with 
        an exponential reparametrization
        :type reparametrization: Function MultiIndex to List-of-MultiIndex, or
            simply True, which corresponds to exponential reparametrization
        '''
        #if C<1:
        #    raise ValueError('Choose C>=1')
        self.C=C
        if isinstance(domain,AbstractProbabilityDistribution):
            self.probability_distribution = domain
        elif domain:
            if swutil.validation.Iterable.valid(domain[0]):
                self.probability_distribution = ProbabilityDistribution(interval = domain[0])
                for dom in domain[1:]:
                    self.probability_distribution = self.probability_distribution * ProbabilityDistribution(interval = dom)
            else:
                self.probability_distribution = ProbabilityDistribution(interval = domain)
        else:
            self.probability_distribution = ProbabilityDistribution()**n
        if swutil.validation.String.valid(sampler) and n_acc:
            self.sampler = lambda mi: sampler
        elif swutil.validation.Instance(np.ndarray).valid(sampler) and n_acc:
            if self.n_acc>1:
                raise ValueArray("Don't currently support predefined samples with multi-index regression")
            self.sampler = lambda mi: sampler[...,mi]
        else:
            self.sampler = functools.partial(sampler,mi) if n_acc else sampler
        self.n_acc = n_acc
        self.warnings=warnings
        if n is None:
            self.n = self.probability_distribution.get_c_var()
        else:
            self.n = n
        self.bundled_dims = lambda dim: dim >= self.n_acc
        self.function = function
        if self.n_acc>0:
            def T(mi_acc): 
                if np.isinf(self.n_acc): # prevent user from accidentally modifying mi_acc
                    mi = copy.deepcopy(mi_acc)
                else:
                    mi = mi_acc.full_tuple(c_dim=self.n_acc)
                return VFunction(lambda X: self.function(mi,X))
            mdT=MixedDifferences(T)
        get_spa = lambda mi_acc: SinglelevelPolynomialApproximator(
            function = mdT(mi_acc) if self.n_acc else self.function,
            probability_distribution = copy.deepcopy(self.probability_distribution),
            C = self.C,
            sampler = self.sampler(mi) if self.n_acc else self.sampler,
            warnings = self.warnings,
        )    
        self.spas = DefaultDict(default=get_spa)
        self.reparametrization=reparametrization
        self.work_per_sample = None
     
    def plot_samples(self,mi_acc=None,ax=None,kwargs=None):
        if mi_acc is None and self.n_acc>0:
            raise ValueError('Must specify which delta to plot')
        if not isinstance(mi_acc,MultiIndex):
            mi_acc = MultiIndex(mi_acc)
        if mi_acc not in self.spas:
            raise ValueError('Only have samples with accuracy parameters {}'.format(list(self.spas.keys())))
        return self.spas[mi_acc].plot_samples(ax=ax,kwargs=kwargs)

    def get_samples(self,mi_acc=None):
        if mi_acc is not None and not isinstance(mi_acc,MultiIndex):
            mi_acc = MultiIndex(mi_acc)
            if mi_acc not in self.spas:
                raise ValueError('Only have samples with accuracy parameters {}'.format(list(self.spas.keys())))
        if mi_acc is None:
            if self.n_acc>0:
                return {mi_acc:(spa.X,spa.Y) for spa in self.spas}
            else:
                return self.spas[MultiIndex()].X,self.spas[MultiIndex()].Y
                 
    #def get_active_dims(self):
    #    return set.union(*[spa.get_active_dims() for spa in self.spas.values()]) 
    
    def get_approximation(self,mi_acc=None):
        '''
        Returns polynomial approximation
        
        :return: Polynomial approximation of :math:`f\colon [a,b]^d\to\mathbb{R}`
        :rtype: Function
        '''
        if mi_acc is not None:
            if not isinstance(mi_acc,MultiIndex):
                mi_acc = MultiIndex(mi_acc)
            if mi_acc not in self.spas:
                raise ValueError('Only have approximation of Deltas with accuracy parameters {}'.format(list(self.spas.keys())))
            return self.spas[mi_acc].get_approximation()
        else:
            return sum([self.spas[mi_acc].get_approximation() for mi_acc in self.spas])
           
    def update_approximation(self, mis):
        r'''
        Expands polynomial approximation.
        
        Converts list of multi-indices into part that describes polynomial basis
        and part that describes accuracy of samples that are being used, then 
        extend polynomial approximation of self.functon based on this information.
        
        
        :param mis: Multi-indices 
        :return: work and contribution associated to mis
        :rtype: (work,contribution) where work is real number and contribution is dictionary mis>reals
        '''
        bundles=indices.get_bundles(mis, self.bundled_dims)
        work = 0
        contributions=dict()
        for bundle in bundles:
            mis_pols,mi_acc= self._handle_mis(bundle)
            (new_work,_) = self.spas[mi_acc].update_approximation(self._pols_from_mis(mis_pols))
            work+=new_work
            #if work>0:
            pa = self.spas[mi_acc].get_approximation()
            contributions.update({mi_acc+mi.shifted(self.n_acc): pa.norm(self._pols_from_mi(mi)) for mi in mis_pols})
        return work, contributions
    
    def reset(self):
        '''
        Delete all stored samples
        '''
        self.__init__(function=self.function, n_acc=self.n_acc, n = self.n , 
            warnings = self.warnings,domain=self.probability_distribution,C=self.C,sampler=self.sampler,reparametrization = self.reparametrization)
    
    def estimated_work(self, mis,multibundles=False):          
        '''
        Return number of samples that would be generated if instance were 
        extended with given multi-index set.
        
        :param mis: (List of) MultiIndex
        :param multibundles: Allow for multi-index sets that contain multiple different accuracies
        :return: Number of new sampls
        '''
        if multibundles:
            bundles = indices.get_bundles(mis,self.bundled_dims)
        else:
            bundles = [mis]
        work = 0
        for bundle in bundles:
            mis_pols, mi_acc = self._handle_mis(bundle)
            work += (self.spas[mi_acc].estimated_work(self._pols_from_mis(mis_pols))
                     *(self.work_per_sample(mi_acc) if self.work_per_sample else 1))
        return work

    def _handle_mis(self, mis):
        mis_pols = [mi.shifted(-self.n_acc) for mi in mis]
        mi_acc = mis[0].mod(self.bundled_dims)
        return mis_pols, mi_acc
     
    def _pols_from_mi(self, mi):
        '''
        Convert multi-index to corresponding polynomials
        
        :param mi: Multi-index
        :return: List of polynomials corresponding to mi
        '''
        if self.reparametrization is True:
            if mi == MultiIndex():
                    return [mi]
            else:
                univariate_entries = []
                for dimension in mi.active_dims():
                    init_range = 2 ** (mi[dimension])-1
                    end_range = 2 ** (mi[dimension]+1)-1
                    univariate_entries.append(range(init_range, end_range))
                return cartesian_product(univariate_entries, mi.active_dims())
        elif self.reparametrization is False:
            return [mi]
        else:
            return self.reparametrization(mi)
            
    def _pols_from_mis(self,mis):
        '''
        Convert multi-indices to corresponding polynomials
        
        :param mis: Multi-indices
        :return: List of polynomials corresponding to mis
        '''
        if self.reparametrization:
            pols=[]
            for mi in mis:
                pols+= self._pols_from_mi(mi)
            return pols
        else:
            return mis

    def _get_info(self):
        return [spa._info for spa in self.spas.values()]

    def accuracy(self):
        try:
            active_spas = [spa for spa in self.spas.values() if spa.polynomial_space.basis]
            recent_spa = active_spas[-1]
            remainder = sum([spa.remainder for spa in active_spas]) + (recent_spa.get_approximation().norm() + recent_spa.remainder if self.n_acc>0 else 0)
            have = sum([spa.get_approximation().norm() for spa in active_spas])
            return remainder/(have+remainder)
        except IndexError:
            return np.inf
        
class SinglelevelPolynomialApproximator:
    '''
    Maintains polynomial approximation of given function on :math:`[a,b]^d`.

    Do not use. Even for single-level approximation it is more convenient to use
    with the general PolynomialApproximator
    '''
    def __init__(self, function, probability_distribution, C=2,sampler='optimal',warnings=False):
        ''' 
        :param function: Function that is approximated. Needs to support __call__
        :param probability_distribution: Probability distribution used for approximation
        :type probability_distribution: ProbabilityDistribution
        :param C: Multiplier for number of samples used for reconstruction
        :type C: Positive real
        '''
        if swutil.validation.String.valid(sampler) and sampler != 'optimal' and sampler != 'arcsine': 
            raise ValueError('Sampling strategy not supported')
        self.keep = True
        self.sampler=sampler
        self.C = C
        self.function = function
        polynomial_space = PolynomialSpace(probability_distribution,warnings= warnings)
        self.polynomial_space = polynomial_space
        self.X = np.zeros((0, self.polynomial_space.get_c_var()))
        self.Y = np.zeros((0, 1))
        self.W = np.zeros((0, 1))
        self._approximation = PolynomialApproximation(self.polynomial_space)
        self._recompute_approximation = True
        self.remainder = np.inf;
       
    def c_samples_per_polynomial(self,dim):
        return max(3,math.ceil(self.C*np.log2(dim+1)))

    def c_samples_from_c_pols(self,dim):
        return dim*self.c_samples_per_polynomial(dim)

    def estimated_work(self,pols):
        '''
        Return number of new samples required to determine polynomial 
        coefficients onto given polynomial subspace
        
        :param pols 
        :return: Number of required samples
        '''
        # if all(pol in self.polynomial_space.basis for pol in pols):
        #     return 0
        c_pols = len(pols)
        c_samples = self.c_samples_from_c_pols(c_pols)
        if self.keep:
            return int(max(c_samples - self.X.shape[0],0))
        else:
            return int(c_samples) 
        
    def update_approximation(self, pols):
        '''
        Get near-optimal projection onto subspace described by mis
        
        :param mis: List of multi-indices describing polynomial span to be added
        :return: Time to compute new samples
        '''
        self._recompute_approximation=True
        old_approximation = copy.deepcopy(self._approximation)
        old_basis = self.polynomial_space.basis
        c_samples = self.estimated_work(pols) # this must come before next call
        self.polynomial_space.set_basis(pols)
        too_few_samples = False
        if self.X.shape[1] < self.polynomial_space.get_c_var():
            c_samples += self.X.shape[0]  # contrary to previous belief we cannot keep samples, thus need to add the number of old ones
            self.X = np.zeros((0, self.polynomial_space.get_c_var()))
            self.Y = np.zeros((0, 1))
            self.W = np.zeros((0, 1))
        if swutil.validation.String.valid(self.sampler):
            if self.sampler == 'arcsine':
                (Xnew,Wnew) = samples.importance_samples(self.polynomial_space.probability_distribution, c_samples, 'arcsine')
            elif self.sampler == 'optimal': 
                (Xnew,Wnew) = samples.samples_per_polynomial(self.polynomial_space,old_basis, pols, self.c_samples_per_polynomial)
            tic = timeit.default_timer()
            Ynew = self.function(Xnew).reshape(-1,1)
            work_model = timeit.default_timer() - tic
        elif swutil.validation.Instance(np.ndarray).valid(self.sampler):
            too_few_samples = len(self.sampler)<c_samples
            if too_few_samples:
                warnings.warn("Need {} more samples".format(c_samples-len(self.sampler)))
                c_samples = len(self.sampler)
                self.polynomial_space.set_basis(old_basis)
                if c_samples:
                    N = len(pols)
                    while c_samples < self.estimated_work(pols[:N]) and N>0:
                        N -= 1 
                    self.polynomial_space.set_basis(pols[:N])
            (samplesnew,self.sampler) = self.sampler[:c_samples],self.sampler[c_samples:]
            (Xnew,Ynew) = samplesnew[:,:-1],samplesnew[:,[-1]]
            Wnew = np.ones((samplesnew.shape[0],1))
            for dim in range(Xnew.shape[1]):
                interval = self.polynomial_space.probability_distribution.ups[dim].interval
                Wnew *= np.pi * np.sqrt((Xnew[:,[dim]] - interval[0]) * (interval[1] - Xnew[:,[dim]]))
            work_model = 0
        else:
            tic = timeit.default_timer()
            (Xnew,Ynew) = self.sampler(c_samples)
            work_model = timeit.default_timer() - tic
            Wnew = np.ones((c_samples, 1))
            for dim in range(Xnew.shape[1]):
                interval = self.polynomial_space.probability_distribution.ups[dim].interval
                Wnew *= np.pi * np.sqrt((Xnew[:,[dim]] - interval[0]) * (interval[1] - Xnew[:,[dim]]))
        if Ynew.shape[0] != Xnew.shape[0]:
            raise ValueError('Function must return as many outputs values as it is given inputs')
        self.X = np.concatenate((self.X,Xnew),axis=0)
        self.W = np.concatenate([self.W, Wnew.reshape(-1,1)])  
        self.Y = np.concatenate([self.Y, Ynew])
        if too_few_samples:
            app = self.get_approximation()
            if not app.coefficients:
                app.coefficients = RFunction()
            for pol in pols:
                if pol not in app.coefficients:
                    app.coefficients[pol] = 0
        if Xnew.shape[0]>0:
            self.remainder = np.sqrt(np.sum(Wnew/np.sum(Wnew)*(old_approximation(Xnew).reshape([-1,1])-Ynew)**2)) 
        return work_model,self.get_contributions()
        
    def get_contributions(self):
        pa=self.get_approximation()
        return {pol:pa.norm([pol]) for pol in self.polynomial_space.basis}
    
    def get_approximation(self):
        if self._recompute_approximation:
            old_err_settings = np.seterr(all='raise')
            while self._recompute_approximation:
                try:
                    (coeffs,self._info) = self.polynomial_space.weighted_least_squares(X = self.X, Y = self.Y, W=self.W)
                    self._approximation =  PolynomialApproximation(self.polynomial_space,coeffs)
                    self._recompute_approximation = False
                except LinAlgError:
                    warnings.warn('Singular matrix, doubling number of samples')
                    c_samples = len(self.X)
                    (Xnew,Wnew) = samples.importance_samples(self.polynomial_space.probability_distribution, c_samples, 'arcsine')
                    self.X = np.concatenate((self.X, Xnew), axis=0)  
                    self.W = np.concatenate([self.W, Wnew])  
                    Ynew=self.function(Xnew).reshape(-1,1)
                    if not Ynew.shape[0]==Xnew.shape[0]:
                        raise ValueError('Function must return as many outputs values as it is given inputs')
                    self.Y = np.concatenate([self.Y, Ynew])
            np.seterr(**old_err_settings)
        return self._approximation
        
    def plot_samples(self,ax=None,kwargs=None):
        '''
        Scatter plot of stored samples.
        '''
        kwargs = kwargs or {}
        if self.polynomial_space.get_c_var() == 1:
            order = np.argsort(self.X, axis=0).reshape(-1)
            if ax is None:
                fig = plt.figure()
                ax = fig.gca()
            ax.scatter(self.X[order, :], self.Y[order, :],**kwargs)
        elif self.polynomial_space.get_c_var() == 2:
            if ax is None:
                fig = plt.figure()
                ax = fig.gca(projection='3d')
            ax.scatter(self.X[:, 0], self.X[:, 1], self.Y,**kwargs)
        return ax

class PolynomialApproximation:
    def __init__(self,ps,coefficients=None,basis = None):
        '''
        :param ps: Polynomial subspace in which approximation lives or ProbabilityDistribution that defines the domain of the approximations
        :type ps: PolynomialSpace or ProbabilityDistribution or 
        :param coefficients: Coefficients with respect to basis determined by
            :code:`basis `
        :type coefficients: Dictionary ps.basis->reals
        '''
        if isinstance(ps,PolynomialSpace):
            self.polynomial_space = ps
            if basis is not None:
                self.polynomial_space.set_basis(basis)
        else:
            self.polynomial_space = PolynomialSpace(ps)
            if basis is None:
                if hasattr(coefficients,'keys'):
                    self.polynomial_space.set_basis(coefficients.keys())
                else:
                    raise ValueError('Must specfiy either polynomial space or probability space and basis or probability space and coefficients dictionary whose keys determine basis')
        self.set_coefficients(coefficients)
        
    def set_coefficients(self,coefficients):
        if hasattr(coefficients,'keys'):
            if set(coefficients.keys()) != set(self.polynomial_space.basis):
                raise ValueError('Coefficients do not match polynomial basis')
            self.coefficients=RFunction(coefficients)
        elif coefficients is not None:
            t=dict()
            if len(self.polynomial_space.basis)!=len(coefficients):
                raise ValueError('Number of coefficients does not match polynomial subspace dimension')
            for i,p in enumerate(self.polynomial_space.basis):
                t[p]=coefficients[i]
            self.coefficients=RFunction(t)
        else:
            self.coefficients = None
        
    def _sum_sq_coeff(self, pols=None):
        '''
        Return sum of squared coefficients of given polynomials
        
        :param mi: Multi-index
        :return: Sum of squares of coefficients corresponding to mi
        '''
        if pols is None:
            pols=self.polynomial_space.basis
        return sum([self.coefficients[pol] ** 2 for pol in pols])  
        
    def __call__(self, X,derivative=None):
        '''
        Return approximation at specified locations.
        
        :param X: Locations of evaluations
        :return: Values of approximation at specified locations
        '''
        if self.coefficients:
            T = self.polynomial_space.evaluate_basis(X,derivative=derivative)
            return T.dot(np.array([self.coefficients[pol] for pol in self.polynomial_space.basis]).reshape(-1))
        else:
            return np.zeros(X.shape[0])
    
    def norm(self, pols=None):
        '''
        Return norm of projection onto polynomial subspace.
        
        :param mi: Multi-index
        :return: Norm of projection onto polynomial space described by mi
        '''
        if pols is None:
            pols=self.polynomial_space.basis
        return np.sqrt(self._sum_sq_coeff(pols))
    
    def __radd__(self, other):
        if other == 0:
            return copy.deepcopy(self)
        else:
            return self.__add__(other)
    
    def __add__(self, other):
        '''
        Add two polynomial approximations. 
        
        :rtype: Function
        '''
        new_coefficients=self.coefficients+other.coefficients
        new_ps=copy.deepcopy(self.polynomial_space)
        new_ps.basis=list(new_coefficients.keys())
        return PolynomialApproximation(ps=new_ps,coefficients=new_coefficients)
        
    def __rmul__(self,other):
        return self.__mul__(other) 
    
    def __mul__(self,other):
        new = copy.deepcopy(self)
        new.coefficients = other*self.coefficients
        return new
    
    def plot(self,ax = None):
        '''
        Plot polynomial approximation.
        '''
        if self.polynomial_space.get_c_var() == 1:
            X = self.polynomial_space.probability_distribution.get_range()
            Z = self(X)
            ax = ax or plt.figure().gca()
            ax.plot(X, Z)
        elif self.polynomial_space.get_c_var() == 2:
            X, Y = self.polynomial_space.probability_distribution.get_range()
            Z = grid_evaluation(X, Y, self)
            ax = ax or plt.figure().gca(projection='3d')
            ax.plot_surface(X, Y, Z,alpha=0.5)
        else:
            raise ValueError('Cannot plot {}-dimensional function'.format(self.polynomial_space.get_c_var()))
        return ax
        



from abc import abstractmethod, ABC
import numpy as np
from numpy import meshgrid
from swutil.validation import Integer, Positive, validate_args

class AbstractProbabilityDistribution(ABC):
    '''
    Probability spaces over subsets of Euclidean space
    '''
    
    @abstractmethod
    def lebesgue_density(self, X): 
        '''
        Return lebesgue density of measure at given locations
        '''
        pass
    
    @abstractmethod
    def get_c_var(self): 
        '''
        Return dimension of domain = number of variables
        '''
        pass 

    @abstractmethod
    def get_range(self, N):
        '''
        Return mesh of points within domain
        '''
        pass

class ProbabilityDistribution(AbstractProbabilityDistribution):
    def __init__(self, measure='u', interval=None):
        if measure  not in ['u', 'c', 'h','t']:
            raise ValueError('Measure not supported')
        else:
            self.measure = measure
        if measure == 'h':
            interval = interval or 1
            self.interval = (None,float(interval)) # replace None by center in the future
        else:
            interval = interval or (-1,1)
            self.interval = (float(interval[0]), float(interval[1]))
        
    def lebesgue_density(self, X):
        if self.measure == 'u':
            return np.ones((X.shape[0], 1)) / (self.interval[1] - self.interval[0])
        elif self.measure == 'c':
            return 1 / (np.pi * np.sqrt((X - self.interval[0]) * (self.interval[1] - X)))
        elif self.measure == 'h':
            return np.exp(-(X ** 2.) / (2.*self.interval[1])) / np.sqrt(2 *self.interval[1]*np.pi)
        
    def get_c_var(self):
        return 1
    
    def get_range(self, N=200,L=None):
        if self.measure in ['u', 'c']:
            interval = self.interval
            L = interval[1] - interval[0]
            X = np.linspace(interval[0] + L / N, interval[1] - L / N, N)
        else:
            L = L or self.interval[1] 
            X = np.linspace(-L, L, N)
        return X.reshape((-1, 1))
    
    def __mul__(self,other):
        if isinstance(other,ProductProbabilityDistribution):
            return ProductProbabilityDistribution([self]+other.ups)
        elif isinstance(other,ProbabilityDistribution):
            return ProductProbabilityDistribution([self,other])
        else:
            raise TypeError()
        
    @validate_args(warnings=False)
    def __pow__(self,other:Positive&Integer):
        return ProductProbabilityDistribution([self]*other)
        

class ProductProbabilityDistribution(AbstractProbabilityDistribution):
    def __init__(self,univariate_probability_distributions):
        self.ups=univariate_probability_distributions
    
    def __mul__(self,other):
        if isinstance(other,ProductProbabilityDistribution):
            return ProductProbabilityDistribution(self.ups+other.ups)
        elif isinstance(other,ProbabilityDistribution):
            return ProductProbabilityDistribution(self.ups+[other])
        else:
            raise TypeError()
    
    @validate_args(warnings=False)
    def __pow__(self,other:Positive&Integer):
        return ProductProbabilityDistribution(self.ups*other)

    def lebesgue_density(self,X):
        D=np.ones((X.shape[0],1))
        for dim in range(self.get_c_var()):
            D*=self.ups[dim].lebesgue_density(X[:,[dim]])
        return D
        
    def get_c_var(self):
        return len(self.ups)
    
    def get_range(self, N=30,L=1):
        if self.get_c_var() == 1:
            if self.ups[0].measure in ['u', 'c']:
                interval = self.ups[0].interval
                L = interval[1] - interval[0]
                X = np.linspace(interval[0] + L / N, interval[1] - L / N, N)
            else:
                X = np.linspace(-L, L, N)
            return X.reshape((-1, 1))
        elif self.get_c_var() == 2:
            T = np.zeros((N, 2))
            for i in [0, 1]:
                if self.ups[i].measure in ['u', 'c']:
                    interval = self.ups[i].interval
                    L = interval[1] - interval[0]
                    T[:, i] = np.linspace(interval[0] + L / N, interval[1] - L / N, N) 
                else:
                    T[:, i] = np.linspace(-L, L, N)
            X, Y = meshgrid(T[:, 0], T[:, 1])
            return (X, Y)
        
        



from __future__ import division #WHY DOES THIS NOT WORK????????????????????????
import numpy as np
import scipy.stats
import itertools
def forward_stepping(particles,weights,h,potential):
    N_steps = int(1. / h)
    d=particles.shape[0]
    h = 1. / N_steps
    N_particles = particles.shape[1]
    particle_history=np.empty((d,N_particles,N_steps+1))
    particle_history[:,:,0]=particles
    for step in range(N_steps):
        for (i, p1) in enumerate(particle_history[:,:,step].T):
            drift=np.zeros((d,))
            for (j, p2) in enumerate(particle_history[:,:,step].T):
                if i != j:
                    drift+= float(weights[j])/ N_particles * h * potential(p1,p2)
            particle_history[:,i,step+1] =p1 + drift
    return particle_history
        
def backward_stepping(particles,weights,h,d_potential):
    N_steps = int(1. / h)
    d=particles.shape[0]
    h = 1. / N_steps
    N_particles = particles.shape[1]
    particle_history=np.empty((d,N_particles,N_steps+1))
    particle_history[:,:,0]=particles
    for step in range(N_steps):
        for (i, p1) in enumerate(particle_history[:,:,step].T):
            drift=np.zeros((d,))
            for (j, p2) in enumerate(particle_history[:,:,step].T):
                if i != j:
                    drift+= -(p1 *float(weights[j]))/N_particles*h*d_potential(p1,p2,1)
                    drift+=  -(p2 *float(weights[i]))/N_particles*h*d_potential(p2,p1,2)       
            particle_history[:,i,step+1] = p1 +drift
    return particle_history
        
def update_control(control,particle_history,costate_history,alpha,potential,rho):
    N_steps=particle_history.shape[2]-1
    N_particles=particle_history.shape[1]
    d=particle_history.shape[0]
    new_control=control.copy()
    new_control[0]=control[0]+rho*(-alpha/control[0]**2-1/N_particles*np.sum(costate_history[:,:,-1]*particle_history[:,:,0]))
    new_control[1:]=control[1:]+15*rho*(-1/N_particles*np.sum(costate_history[:,:,-1],axis=1))
    #control[1]=max([control[1],0])
    return new_control

def iteration(c_particles,c_steps,c_iter,alpha,d,rho,potential,d_potential,control,random=False):
    print(c_particles,c_steps)
    h=1./c_steps
    
    new_control=control
    if random:
        np.random.seed(10)
        normal_discretization = np.random.multivariate_normal(np.zeros((d,)),np.eye(d),size=(c_particles,)).T
    else:
        normal_discretization = np.array(list(itertools.product(*([scipy.stats.norm.ppf(np.linspace(0,1,c_particles+2)[1:-1])]*d)))).T
    N_particles=normal_discretization.shape[1]
    weights=np.ones((N_particles,))
    for __ in range(c_iter):
        control=new_control
        init_particles=np.tile(control[1:,None],(1,N_particles))+control[0]*normal_discretization
        particle_history=forward_stepping(init_particles,weights,h,potential)
        costate_history=backward_stepping(-particle_history[:,:,-1],weights,h,d_potential)
        new_control=update_control(control,particle_history,costate_history,alpha,potential,rho)
        #print(new_control)
    return control
        
if __name__=='__main__':
    np.seterr(all='raise')
    N=4
    c_steps=20
    c_iter=100
    alpha=1
    rho=-0.05
    power_reject=1
    power_attract=1
    d=2
    def potential(p1,p2,power):
        return (p2 - p1) * np.linalg.norm(p2 - p1)**(power-1)
    def d_potential(p1,p2,power,k):
        return (-1)**k*(np.linalg.norm(p2-p1)**(power-1)+(power-1)*np.power(p2-p1,2)*np.linalg.norm(p2-p1)**(power-3))
        #return (-1)**k*power*np.linalg.norm(p2-p1)**(power-1)
    def total_potential(p1,p2):
        return -potential(p1,p2,power_reject)#+potential(p1,p2,power_attract)
    def total_d_potential(p1,p2,k):
        return -d_potential(p1,p2,power_reject,k)#+d_potential(p1,p2,power_attract,k)
    control=np.array([0.1]+d*[1])
    print(iteration(N,c_steps,c_iter,alpha,1,rho,total_potential,total_d_potential,control))
    